/*
 *     Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 *     Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package edu.ucsc.its.temerity.core

import io.ktor.client.HttpClient
import io.ktor.client.HttpClientConfig
import io.ktor.client.engine.HttpClientEngineConfig
import io.ktor.client.engine.HttpClientEngineFactory
import io.ktor.client.plugins.logging.LoggingConfig
import kotlin.time.Duration

/**
 * Configuration for the Temerity client.
 * Adapted from the tmdb-api project: https://github.com/MoviebaseApp/tmdb-kotlin/raw/refs/heads/main/tmdb-api/src/commonMain/kotlin/app/moviebase/tmdb/TmdbClientConfig.kt
 *
 * @property serviceUrl Specifies URL of platform instance to make API calls to
 * @property serviceToken Specifies the authorization token to use for API calls
 * @property optLoggingEnabled Specifies whether logging should be enabled
 * @property optDebugEnabled Specifies whether debugging mode should be enabled
 * @property optSupportKtxNotebook Specifies whether Kotlin notebook support should be enabled
 *
 * @property optExpectSuccess Specifies whether to expect successful responses by default
 * @property optUseWebTimeout Specifies whether to use a timeout  other than default for web requests
 * @property optWebTimeoutDuration Specifies the timeout [kotlin.time.Duration] to use when making web requests
 * @property optCacheTimeoutDuration Specifies the timeout [kotlin.time.Duration] to use when storing cache entries. Affects keep-alive time for cached instance data like user role type fields.
 * @property optThreadCount Specifies the number of threads to use for each client instance. Defaults to 2 at a minimum
 * @property optEnableCompression Specifies whether to turn on compression for HTTP requests. Enabled by default.
 */
@TemDsl
public class TemClientConfig {

  public var serviceUrl: String? = null
  public var serviceToken: String? = null
  public var optLoggingEnabled: Boolean = true
  public var optDebugEnabled: Boolean = false
  public var optSupportKtxNotebook: Boolean = false

  public var optExpectSuccess: Boolean = true
  public var optUseWebTimeout: Boolean = false

  @Suppress("MemberVisibilityCanBePrivate")
  public var optWebTimeoutDuration: Duration? = null

  @Suppress("MemberVisibilityCanBePrivate")
  // Default cache timeout is set to 15 minutes if this is left unspecified; @see [edu.ucsc.its.temerity.core.Temerity] init block
  public var optCacheTimeoutDuration: Duration? = null

  public var optThreadCount: Int? = null

  public var optEnableCompression: Boolean = true

  /**
   * Configures the HttpClient with the provided block.
   * @param () -> HttpClient The block to configure the HttpClient.
   */
  internal var optHttpClientConfigBlock: (HttpClientConfig<*>.() -> Unit)? = null
  internal var optHttpClientBuilder: (() -> HttpClient)? = null
  internal var optHttpClientLoggingBlock: (LoggingConfig.() -> Unit)? = null

  public fun logging(block: LoggingConfig.() -> Unit) {
    optHttpClientLoggingBlock = block
  }

  /**
   * Set custom HttpClient configuration for the default HttpClient.
   */
  public fun httpClient(block: HttpClientConfig<*>.() -> Unit) {
    optHttpClientConfigBlock = block
  }

  /**
   * Creates an custom [HttpClient] with the specified [io.ktor.client.engine.HttpClientEngineFactory] and optional [block] configuration.
   * Note that the Temerity config will be added afterward.
   */
  public fun <T : HttpClientEngineConfig> httpClient(
    engineFactory: HttpClientEngineFactory<T>,
    block: HttpClientConfig<T>.() -> Unit = {},
  ) {
    optHttpClientBuilder = {
      HttpClient(engineFactory, block)
    }
  }

  public companion object {
    internal fun withToken(serviceToken: String): TemClientConfig = TemClientConfig().apply {
      this.serviceToken = serviceToken
    }
  }
}
