/*
 *     Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 *     Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package edu.ucsc.its.temerity.extensions.coroutines

import edu.ucsc.its.temerity.core.Temerity.Companion.DEFAULT_MINIMUM_THREAD_COUNT
import io.ktor.util.SilentSupervisor
import kotlinx.coroutines.CoroutineDispatcher
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Job
import kotlinx.coroutines.SupervisorJob
import kotlin.coroutines.CoroutineContext

internal fun createJobScope(coroutineContext: CoroutineContext, allowIndependentFailure: Boolean = false): CoroutineScope {
  val parentJob = if (allowIndependentFailure) SupervisorJob() else Job()
  return CoroutineScope(coroutineContext + parentJob)
}

internal fun createLibraryScope(dispatcher: CoroutineDispatcher): CoroutineScope = CoroutineScope(dispatcher + SilentSupervisor())

internal fun availableThreads(maximumThreadCount: Int? = null): Int {
  val availableThreadCount = Runtime.getRuntime().availableProcessors().plus(1) // +1 to maintain full utilization in case one coroutine blocks. Never less than 2 (DEFAULT_MINIMUM_THREAD_COUNT). See: https://github.com/Kotlin/kotlinx.coroutines/issues/261
  with(maximumThreadCount) {
    return when {
      this == null -> availableThreadCount
      this > availableThreadCount -> {
        // TODO: Log warning that the configured thread count is higher than the number of available threads
        availableThreadCount
      }
      this < DEFAULT_MINIMUM_THREAD_COUNT -> {
        // TODO: Log warning that the configured thread count is lower than the default minimum thread pool size
        DEFAULT_MINIMUM_THREAD_COUNT
      }
      this < availableThreadCount -> {
        // TODO: Log warning that the configured thread count is lower than the optimal thread pool size
        this
      }
      else -> availableThreadCount
    }
  }
}
