/*
 *     Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 *     Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package edu.ucsc.its.temerity.extensions.log

import co.touchlab.kermit.DefaultFormatter
import co.touchlab.kermit.LogWriter
import co.touchlab.kermit.Message
import co.touchlab.kermit.MessageStringFormatter
import co.touchlab.kermit.Severity
import co.touchlab.kermit.Tag
import kotlinx.io.buffered
import kotlinx.io.files.FileSystem
import kotlinx.io.files.Path
import kotlinx.io.writeString

internal expect fun fileSystem(): FileSystem

internal class FilesystemLogWriter internal constructor(
  private val logPath: String,
  private val logRoller: LogRoller? = null,
  private val formatter: MessageStringFormatter = DefaultFormatter,
) : LogWriter() {

  // Not called since we are "Context Aware"
  override fun log(severity: Severity, message: String, tag: String, throwable: Throwable?) {
    val fileSystem = fileSystem()
    val kotlinxIoPath = Path(logPath)

    logRoller?.rollLogs(kotlinxIoPath, fileSystem)

    val sink = fileSystem.sink(kotlinxIoPath, append = true).buffered()

    with(sink) {
      writeString(formatter.formatMessage(severity, Tag(tag), Message(message)))
      writeString("\n")
      throwable?.let {
        writeString(it.stackTraceToString())
      }
      flush()
    }
  }

  companion object {
    operator fun invoke(block: Builder.() -> Unit) = with(FilesystemLogWriterBuilder()) {
      block(this)
      build()
    }
  }

  interface Builder {
    fun rollLogAtSize(size: Long): Builder
    fun logPath(path: String): Builder
    fun build(): FilesystemLogWriter
  }

  class FilesystemLogWriterBuilder : Builder {
    private var maxFileSize: Long? = null
    private var logPath: String? = null

    override fun rollLogAtSize(size: Long): Builder {
      maxFileSize = size
      return this
    }

    override fun logPath(path: String): Builder {
      logPath = path
      return this
    }

    override fun build(): FilesystemLogWriter {
      if (logPath == null) throw NullPointerException("Invalid / missing log path")

      // Can you resist the urge to Rick-roll the logs?
      val rick = maxFileSize?.let {
        FileSizeLogRoller(logPath!!, maxFileSize!!)
      }

      return FilesystemLogWriter(logPath!!, rick)
    }
  }
}
