/*
 *     Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 *     Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package edu.ucsc.its.temerity.extensions.log

import co.touchlab.kermit.LogWriter
import co.touchlab.kermit.Logger
import co.touchlab.kermit.Severity
import co.touchlab.kermit.StaticConfig

/*
 * An overload for the Kermit logger that allows for more concise and readable configuration code.
 * Adapted from: https://github.com/psh/KermitExt/blob/060cd7fab8a73af4187498af8ed169fa4624bf5a/kermit-config/src/commonMain/kotlin/com/gatebuzz/kermit/ext/Kermit.kt
 */
internal class Kermit {

  companion object {
    @JvmStatic
    fun builder(): Builder = LoggerBuilder()

    operator fun invoke(block: Builder.() -> Unit): Logger = with(LoggerBuilder()) {
      block(this)
      build()
    }
  }

  interface Builder {
    fun tag(tag: String): Builder
    fun minSeverity(severity: Severity): Builder
    fun setLogWriters(vararg logWriter: LogWriter): Builder
    fun addLogWriter(vararg logWriter: LogWriter): Builder
    operator fun LogWriter.unaryPlus()
    fun build(): Logger
  }

  internal class LoggerBuilder : Builder {
    private var tag: String = ""
    private var logWriters: MutableList<LogWriter> = mutableListOf()
    private var minSeverity: Severity = Severity.Verbose

    override fun tag(tag: String): Builder {
      this.tag = tag
      return this
    }

    override fun minSeverity(severity: Severity): Builder {
      this.minSeverity = severity
      return this
    }

    override fun setLogWriters(vararg logWriter: LogWriter): Builder {
      logWriters = logWriter.toMutableList()
      return this
    }

    override operator fun LogWriter.unaryPlus() {
      logWriters.add(this)
    }

    override fun addLogWriter(vararg logWriter: LogWriter): Builder {
      logWriters.addAll(logWriter)
      return this
    }

    override fun build(): Logger {
      if (logWriters.isEmpty()) throw Exception("At least one log writer is needed")

      return Logger(
        StaticConfig(minSeverity, logWriters.toList()),
      )
    }
  }
}
